r"""This module contains a set of global configuration variables for the simulation,
such as the engine to use, and a set of functions to make the configuration more user-friendly."""

import csv

import pandas as pd
import pkg_resources

from omnetpypy import utilities, sim_log, parser
from omnetpypy.backends.simpy_connector import SimPyConnector


class Simulation:
    """
    This class is a container for the simulation configuration parameters.
    It represents a simulation configuration, and it can be used to store the configuration of a repetition
    """

    def __init__(self, engine, seed_set, repetition, metrics, yaml_path, until, log_level, time_unit, output_dir,
                 global_params):
        self.engine = engine
        self.seed_set = seed_set
        self.repetition_idx = repetition
        self.until = until
        self.log_level = log_level
        self.time_unit = time_unit
        self.output_dir = output_dir

        self.global_params = global_params.copy()
        # dictionary of global parameters

        self.rng = utilities.MultiRandom(seeds=seed_set)
        self.connector = None

        # set log level
        sim_log.log_to_console(level=self.log_level)

        self.metrics = metrics
        # list of Metrics

        if engine == "simpy":
            self.connector = SimPyConnector(simulation=self, metrics=metrics, output_dir=output_dir, repetition=repetition)
        else:
            raise NotImplementedError("Engine not implemented")

        # parse topology yaml file
        self.network = parser.parse_topology(yaml_path, self)

    def start(self):
        self.connector.start_simulation(until=self.until)
        collected_data = {}
        # collect metrics
        for metric in self.metrics:
            # dump the remaining samples in memory to the temporary file
            self.connector.dump_metric(metric.name)

            # compute the statistics we need
            filename = f"{self.output_dir}/.{metric.name}_vector_rep{self.repetition_idx}.csv"
            collected_data[metric.name] = utilities.get_metrics_from_csv(metric, filename)

        return collected_data

    def time(self):
        """
        Return the current simulation time.

        Returns
        -------
        int or float
            The current simulation time.

        """
        return self.connector.get_time()

    @property
    def time_unit_factor(self):
        """
        Return the factor to convert the time unit to seconds.

        Returns
        -------
        int or float
            The factor to convert the time unit to seconds.

        """
        return utilities.time_unit_factor(self.time_unit)


class Experiment:
    """
    This class is a container for the experiment configuration parameters.
    It represents an experiment configuration, and it can be used to store the configuration of a set of
    simulated repetitions
    """

    def __init__(self, config_file):

        self.seed_sets = []

        # every repetition has its own seed set of length rngs_per_rep
        # seeds are taken from the seeds.csv file
        # with a single "seeds" column. We open and read the file here
        # but not all the seeds are used, only the first rngs_per_rep * repetitions

        self.config = parser.parse_config(config_file)

        engine = self.config.get("engine", "simpy")
        repetitions = self.config.get("repetitions", 1)
        yaml_path = self.config.get("topology_file", "topology.yaml")
        rngs_per_rep = self.config.get("num_rngs", 1)
        until = self.config.get("simulate_until", None)
        log_level = self.config.get("log_level", "info")
        time_unit = self.config.get("time_unit", "us")

        global_params = self.config.get("global_params", {})

        # set log level
        sim_log.log_to_console(level=log_level)

        # so we iterate over the first rngs_per_rep * repetitions seeds
        # and we store them in a list of lists, where each list of seeds
        # is the seed set for a repetition
        seeds_path = pkg_resources.resource_filename('omnetpypy', 'data/seeds.csv')

        with open(seeds_path, "r") as f:
            reader = csv.reader(f)
            # use the reader sequentially on the first rngs_per_rep * repetitions rows
            for _ in range(repetitions):
                self.seed_sets.append([])
                for _ in range(rngs_per_rep):
                    self.seed_sets[-1].append(int(next(reader)[0]))

        metrics_raw = self.config.get("metrics", [])
        metrics = []

        for metric in metrics_raw:  # set up metric collection
            name = metric.get("name")
            types = metric.get("type")

            all_types = ["vector", "mean", "std", "min", "max", "median", "percentiles", "count", "var"]
            params = {k: False for k in all_types}

            true_params = {k: True for k in types}
            params.update(true_params)

            params["name"] = name

            metrics.append(utilities.FutureMetric(**params))

        output_dir = self.config.get("output_dir", "output")

        self.simulations_params = [(engine, self.seed_sets[i], i, metrics, yaml_path,
                                    until, log_level, time_unit, output_dir, global_params)
                                   for i in range(repetitions)]

    def run_simulations(self):

        import multiprocessing

        max_processes = self.config.get("max_processes", 1)
        # check if max_processes is greater than the number of cpu cores
        if max_processes > 1:
            max_processes = min(multiprocessing.cpu_count(), max_processes)

        # run the simulations in parallel with a ProcessPoolExecutor
        with multiprocessing.Pool(max_processes) as pool:
            collected = pool.map(_start_sim, self.simulations_params)

            # every element of collected is a dictionary of dictionaries with the same keys.
            # we want to merge them into a single dictionary, where the values
            # are lists, except for the "name" key and the "vector" key, because the latter
            # is a pandas DataFrame and we will return a single merged DataFrame

            # first we create a dictionary
            merged = {}

            # now we iterate over the collected data
            for metric in collected[0].keys():
                merged[metric] = {}

            # now we iterate over the collected data
            for data in collected:
                for metric in data.keys():
                    for key, value in data[metric].items():
                        if key != "vector":
                            if key not in merged[metric]:
                                merged[metric][key] = []
                            merged[metric][key].append(value)

            out_dir = self.config.get("output_dir", "output")
            # write the data to csv, two files for each metric: "metric_name.csv" and "metric_name_vector.csv"
            for metric in merged.keys():
                # we write the statistics to a file if merged[metric] is not empty
                if merged[metric]:
                    df = pd.DataFrame(merged[metric])
                    df.to_csv(f"{out_dir}/{metric}.csv", index=False)

            for metric in merged.keys():
                # we open the temporary files and merge them into a single file. We add a "repetition" column
                max_reps = self.config.get("repetitions", 1)
                for i in range(max_reps):
                    if i == 0:
                        with open(f"{out_dir}/.{metric}_vector_rep{i}.csv", "r") as f:
                            # add the repetition column
                            df = pd.read_csv(f)
                            if max_reps > 1:
                                df["repetition"] = i
                            df.to_csv(f"{out_dir}/{metric}_vector.csv", index=False)

                    else:
                        with open(f"{out_dir}/.{metric}_vector_rep{i}.csv", "r") as f:
                            # add the repetition column
                            df = pd.read_csv(f)
                            if max_reps > 1:
                                df["repetition"] = i
                            df.to_csv(f"{out_dir}/{metric}_vector.csv", mode="a", header=False, index=False)
            # remove the temporary files
            import os
            for metric in merged.keys():
                for i in range(self.config.get("repetitions", 1)):
                    os.remove(f"{out_dir}/.{metric}_vector_rep{i}.csv")


def _start_sim(sim_params):
    # defined here to be picklable
    engine, seed_set, repetition, metrics, yaml_path, until, log_level, time_unit, output_dir, global_params = sim_params
    sim = Simulation(engine, seed_set, repetition, metrics, yaml_path, until, log_level, time_unit, output_dir, global_params)
    return sim.start()



"""
This module implements the Connector class, which is an interface to the simulation engine.
Connectors provide a set of functions to interact with the simulation engine,
such as starting and stopping the simulation, adding and removing entities from the simulation,
supporting event generation and management.
"""
from abc import abstractmethod
import abc
import pandas as pd


class Connector(abc.ABC):
    """
    This class is an interface to the simulation engine.
    This is an abstract class, and it is meant to be subclassed by connectors to specific simulation engines.

    Parameters
    ----------
    simulation : Simulation
        The simulation object that describes the simulation configuration.
    """

    def __init__(self, simulation, metrics=None, output_dir=None, repetition=0):
        self.simulation = simulation
        self.metrics = metrics
        self.output_dir = output_dir
        self.repetition = repetition

        # allocate a dict of numpy dataframes indexed by metric name
        if metrics is not None:
            self.metrics_data = {metric.name: pd.DataFrame(columns=["sample", "timestamp"]) for metric in metrics}
            self.metrics_headers = {metric.name: False for metric in metrics}
        else:
            self.metrics_data = {}

    @abstractmethod
    def start_simulation(self, until=None):
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def add_entity(self, entity):
        entity.set_sim_context(self.simulation)

        # check if entity has property sub_modules
        if hasattr(entity, "sub_entities"):
            for _, entity in entity.sub_entities.items():
                self.add_entity(entity)

    @abstractmethod
    def get_time(self):
        raise NotImplementedError("to be implemented by subclasses")

    def random(self):
        """
        Return the MultiRandom random number generator instantiated for the current simulation.

        Returns
        -------
        MultiRandom
            A random number generator that supports multiple independent streams.
        """
        return self.simulation.rng

    @abstractmethod
    def schedule_port_input(self, port, message):
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def schedule_self_message(self, message, entity, at=None, delay=None):
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def is_scheduled(self, message, entity):
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def cancel_scheduled(self, message, entity):
        raise NotImplementedError("to be implemented by subclasses")

    def record_metric(self, metric, value):
        """
        Record a metric value for the current simulation.

        Parameters
        ----------
        metric : str
            The name of the metric to be recorded.
        value : float
            The value of the metric to be recorded.
        """
        if self.metrics is not None:
            # add the value to the metric dataframe by simply appending a new row using loc

            if metric in self.metrics_data:
                self.metrics_data[metric].loc[len(self.metrics_data[metric].index)] = [value, self.get_time()]
            # otherwise we ignore the emitted metric

                # append the metrics to the output file if the dataframes are too big
                if len(self.metrics_data[metric].index) > 1000:
                    self.dump_metric(metric)

        else:
            raise Exception("No metrics have been defined for this simulation")

    def dump_metric(self, metric):
        """
        Dump the metric data to a file.

        Parameters
        ----------
        metric : str
            The name of the metric to be dumped.
        """
        if self.metrics is not None:
            if metric in self.metrics_data:
                if not self.metrics_headers[metric]:
                    self.metrics_data[metric].to_csv(f"{self.output_dir}/.{metric}_vector_rep{self.repetition}.csv", mode="w", header=True, index=False)
                    self.metrics_headers[metric] = True
                else:
                    self.metrics_data[metric].to_csv(f"{self.output_dir}/.{metric}_vector_rep{self.repetition}.csv", mode="a", header=False, index=False)
                self.metrics_data[metric] = pd.DataFrame(columns=["sample", "timestamp"])
        else:
            raise Exception("No metrics have been defined for this simulation")




"""
This module is responsible for parsing the YAML configuration file and returning
the compound module containing the whole simulation topology.
"""
import importlib

import yaml

from omnetpypy.front_end import CompoundModule
from omnetpypy.front_end.channel import Channel


def parse_simple_modules(simple_data):
    module_classes = []
    for module_data in simple_data:
        name = module_data['name']
        package = module_data.get('package', None)
        # Import module dynamically
        # import with importlib
        imported = importlib.import_module(name=package)
        module_class = getattr(imported, name)

        module_classes.append(module_class)
    return module_classes


def parse_channels(channel_data):
    channel_classes = []
    for channel_data in channel_data:
        name = channel_data['name']
        package = channel_data['package']
        # Import channel dynamically if not already imported
        imported = importlib.import_module(name=package)
        channel_class = getattr(imported, name)
        channel_classes.append(channel_class)
    return channel_classes


def parse_compound_modules(compound_data, module_classes, channel_classes):
    compound_modules = []
    next_simple_id = 10100
    next_compound_id = 1111100000
    next_channel_id = 2222200000
    for compound_mod in compound_data:
        name = compound_mod['name']
        ports = [port['name'] for port in compound_mod.get('ports', [])]
        parameters = {'name': name, 'identifier': next_compound_id, 'port_names': ports, 'parent': None}
        kwargs = compound_mod.get('parameters', {})
        parameters.update(kwargs)
        next_compound_id += 1
        submodules = []
        compound_module = CompoundModule(**parameters)
        compound_modules.append(compound_module)
        for submodule_data in compound_mod.get('submodules', []):
            type_name = submodule_data['type']
            # find the submodule class iterating over the module_classes and looking for the class with the same name
            submodule_class = None
            for module_class in module_classes:
                if module_class.__name__ == type_name:
                    submodule_class = module_class
                    break
            if submodule_class is None:
                raise Exception(f"Module class {type_name} not found")

            sub_name = submodule_data['name']
            parameters = submodule_data.get('parameters', {})
            parameters['name'] = sub_name
            parameters['identifier'] = next_simple_id
            next_simple_id += 1
            submodule_instance = submodule_class(**parameters)
            compound_module.add_sub_entity(submodule_instance)
            submodules.append(sub_name)

        for connection_data in compound_mod.get('connections', []):
            # we check whether there is a field name "for i in x to y"
            # if so, we have to create a connection for iteration
            # get the keys of the dictionary and check if one starts with "for"
            keys = connection_data.keys()
            loop_found = False
            for key in keys:
                if key.startswith("for"):
                    # get the range of the iteration
                    bottom = int(key.split(" ")[3])
                    top = int(key.split(" ")[5])
                    # get the name of the variable
                    var_name = key.split(" ")[1]
                    # get the connection data
                    conn_data = connection_data[key]
                    # iterate over the range
                    for i in range(bottom, top + 1):
                        # format conn_data['source'] and conn_data['target'] by replacing the variable name
                        new_data = conn_data.copy()
                        for k, v in new_data.items():
                            if k == "source" or k == "target":
                                new_data[k] = v.replace("{"+var_name+"}", str(i))

                                # there might also occur "{i+l}" or "{i-l}" in the string
                                # we have to replace these as well using pattern matching
                                # we can use the re module for this
                                # we have to import the re module
                                import re
                                # we have to find all occurences of "{i+l}" and "{i-l}" in the string, for any l
                                # we can use the findall method of the re module
                                # we have to define the pattern first
                                pattern = r"{i[+-]\d+}"
                                # then we can use the findall method
                                matches = re.findall(pattern, new_data[k])
                                # we have to iterate over the matches and replace them in the string
                                for match in matches:
                                    # we have to extract the number and the operator from the match
                                    operator = match[2]
                                    number = int(match[3:-1])
                                    # we have to replace the match in the string
                                    if operator == "+":
                                        new_data[k] = new_data[k].replace(match, str(i+number))
                                    else:
                                        new_data[k] = new_data[k].replace(match, str(i-number))

                        # parse the connection
                        next_channel_id = parse_connection(new_data, compound_module, channel_classes,
                                                           next_channel_id)
                        loop_found = True

            if not loop_found:
                next_channel_id = parse_connection(connection_data, compound_module, channel_classes,
                                                   next_channel_id)

    return compound_modules


def parse_connection(connection_data, compound_module, channel_classes, next_channel_id):
    source = connection_data['source']
    target = connection_data['target']
    source_module, source_port = source.split('.')
    target_module, target_port = target.split('.')
    if source_module == "self":
        source_module = compound_module
    else:
        source_module = compound_module.sub_entities[source_module]
    if target_module == "self":
        target_module = compound_module
    else:
        target_module = compound_module.sub_entities[target_module]

    # check if the "type" field is present

    if "type" in connection_data and connection_data["type"] == "forward input":
        source_module.ports[source_port].forward_input(target_module.ports[target_port])

    elif "type" in connection_data and connection_data["type"] == "forward output":
        source_module.ports[source_port].forward_output(target_module.ports[target_port])

    elif "type" in connection_data and connection_data["type"] == "subscribed":
        target_module.ports[target_port].subscribe_to(source_module.ports[source_port])

    elif "type" not in connection_data and "channel" not in connection_data:
        channel = None
        source_module.connect(local_port=source_port, remote_entity=target_module, remote_port=target_port,
                              channel=channel)

    elif "type" not in connection_data and "channel" in connection_data:
        if connection_data["channel"] == "default":
            # in this case we gather the parameters and instantiate a default channel
            parameters = connection_data.get("parameters", {}).copy()
            parameters["name"] = "channel_" + str(next_channel_id)
            parameters["identifier"] = next_channel_id
            next_channel_id += 1
            channel = Channel(**parameters)
        else:
            # in this case we have to instantiate the channel
            channel_classname = connection_data["channel"]
            # find the channel class iterating over the module_classes and looking for the class with the same name
            channel_class = None
            for channel_class in channel_classes:
                if channel_class.__name__ == channel_classname:
                    channel_class = channel_class
                    break
            parameters = connection_data.get("parameters", {})
            parameters["name"] = f"{channel_class}_" + str(next_channel_id)
            parameters["identifier"] = next_channel_id
            next_channel_id += 1
            channel = channel_class(**parameters)

        # the channel is added to the compound module as a sub entity
        compound_module.add_sub_entity(channel)

        source_module.connect(local_port=source_port, remote_entity=target_module, remote_port=target_port,
                              channel=channel)
    else:
        raise Exception(f"Invalid connection: {connection_data}")

    return next_channel_id


# Parse the YAML configuration
def parse_topology(config_file, sim_context):
    with open(config_file, 'r') as file:
        config = yaml.safe_load(file)
        simple_modules = parse_simple_modules(config.get('simple', []))
        channels = parse_channels(config.get('channels', []))
        compound_modules = parse_compound_modules(config.get('network', []), simple_modules, channels)
        if len(compound_modules) == 0 or len(compound_modules) > 1:
            raise Exception("Only one compound module is allowed in the network for the moment")
        sim_context.connector.add_entity(compound_modules[0])

    return compound_modules[0]


def parse_config(config_file):
    with open(config_file, 'r') as file:
        config = yaml.safe_load(file)
    return config



"""This module implements a specific logging API for simulations."""

import logging

__all__ = ["log_to_console", "log_to_file", "set_log_level",
           "remove_console_log", "debug", "info", "warning", "error", "critical"]


class ColorFormatter(logging.Formatter):
    """Class used to format log output color depending on its level.

    DEBUG: blue
    INFO: grey
    WARNING: yellow
    ERROR: red
    CRITICAL: bold red

    Parameters
    ----------
    fmt : str, optional
        The text format to apply to the log entry. Defaults to the message alone.
    """

    grey = '\x1b[37m'
    blue = '\x1b[38;5;39m'
    yellow = '\x1b[38;5;226m'
    red = '\x1b[38;5;196m'
    bold_red = '\x1b[31;1m'
    reset = '\x1b[0m'

    def __init__(self, fmt="%(message)s"):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.blue + self.fmt + self.reset,
            logging.INFO: self.grey + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


logger = logging.getLogger("QI_Logger")
logger.setLevel(logging.DEBUG)


def log_to_console(level=logging.DEBUG):
    """Activate log output to the stdout console.

    Parameters
    ----------
    level : int, optional
        The logging level for the console. Defaults to logging.DEBUG

    Returns
    -------
    :class:`logging.StreamHandler`
        The handler of the log output.
    """
    ch = logging.StreamHandler()
    level = sanitize_log_level(level)
    ch.setLevel(level)
    ch.setFormatter(ColorFormatter())
    remove_console_log()  # to avoid double log on console if user calls this function twice
    logger.addHandler(ch)
    return ch


def log_to_file(filename, level=logging.DEBUG):
    """Activate log output to the specified file.

    Parameters
    ----------
    filename : str
        The name of the output log file. If not present it is created. It is opened in append mode.
    level : int, optional
        The logging level on the file. Defaults to `logging.DEBUG`

    Returns
    -------
    logging.FileHandler
        The handler of the log output.
    """
    fh = logging.FileHandler(filename=filename)
    level = sanitize_log_level(level)
    fh.setLevel(level)
    fh.setFormatter(logging.Formatter(''))
    logger.addHandler(fh)
    return fh


def set_log_level(level):
    """Set the log level on all outputs.

    Parameters
    ----------
    level : int
        The new log level.
    """

    level = sanitize_log_level(level)

    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)


def remove_console_log():
    """Remove the console log handler."""
    h = None
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            h = handler
            break
    if h is not None:
        logger.removeHandler(h)


def sanitize_log_level(level):

    if isinstance(level, str):
        level = level.upper()
        if level == "DEBUG":
            level = logging.DEBUG
        elif level == "INFO":
            level = logging.INFO
        elif level == "WARNING":
            level = logging.WARNING
        elif level == "ERROR":
            level = logging.ERROR
        elif level == "CRITICAL":
            level = logging.CRITICAL
        else:
            raise ValueError(f"Invalid log level: {level}")

    return level


def debug(message, module_id=None, time=None):
    """Log a message with level DEBUG on the logger.

    Parameters
    ----------
    message : str
        The message to be logged. It should not contain the current simulation time because it is already
        present in the log format.
    module_id : int or str or None optional
        If not None, an additional string is added to the log entry, containing the module identifier.
    time : int or float or None, optional
        If not None, an additional string is added to the log entry, containing the current simulation time.
    """
    log_message = ""
    if time is not None:
        if int(time) == time:
            log_message = f"[{int(time)}]::"
        else:
            log_message = f"[{time:.3f}]::"

    log_message += "DEBUG::"
    if module_id is not None:
        log_message += f"MODULE-{module_id}::"
    log_message += (" " + message)
    logger.debug(log_message)


def info(message, module_id=None, time=None):
    """Log a message with level INFO on the logger.

    Parameters
    ----------
    message : str
        The message to be logged. It should not contain the current simulation time because it is already
        present in the log format.
    module_id : int or str or None optional
        If not None, an additional string is added to the log entry, containing the module identifier.
    time : int or float or None, optional
        If not None, an additional string is added to the log entry, containing the current simulation time.
    """
    log_message = ""
    if time is not None:
        log_message = f"[{time}]::"

    log_message += "INFO::"
    if module_id is not None:
        log_message += f"MODULE-{module_id}::"
    log_message += (" " + message)
    logger.info(log_message)


def warning(message, module_id=None, time=None):
    """Log a message with level WARNING on the logger.

    Parameters
    ----------
    message : str
        The message to be logged. It should not contain the current simulation time because it is already
        present in the log format.
    module_id : int or str or None optional
        If not None, an additional string is added to the log entry, containing the module identifier.
    time : int or float or None, optional
        If not None, an additional string is added to the log entry, containing the current simulation time.
    """
    log_message = ""
    if time is not None:
        log_message = f"[{time}]::"

    log_message += "WARNING::"
    if module_id is not None:
        log_message += f"MODULE-{module_id}::"
    log_message += (" " + message)
    logger.warning(log_message)


def error(message, module_id=None, time=None):
    """Log a message with level ERROR on the logger.

    Parameters
    ----------
    message : str
        The message to be logged. It should not contain the current simulation time because it is already
        present in the log format.
    module_id : int or str or None optional
        If not None, an additional string is added to the log entry, containing the module identifier.
    time : int or float or None, optional
        If not None, an additional string is added to the log entry, containing the current simulation time.
    """
    log_message = ""
    if time is not None:
        log_message = f"[{time}]::"

    log_message += "ERROR::"
    if module_id is not None:
        log_message += f"MODULE-{module_id}::"
    log_message += (" " + message)
    logger.error(log_message)


def critical(message, module_id=None, time=None):
    """Log a message with level CRITICAL on the logger.

    Parameters
    ----------
    message : str
        The message to be logged. It should not contain the current simulation time because it is already
        present in the log format.
    module_id : int or str or None optional
        If not None, an additional string is added to the log entry, containing the module identifier.
    time : int or float or None, optional
        If not None, an additional string is added to the log entry, containing the current simulation time.
    """
    log_message = ""
    if time is not None:
        log_message = f"[{time}]::"

    log_message += "CRITICAL::"
    if module_id is not None:
        log_message += f"MODULE-{module_id}::"
    log_message += (" " + message)
    logger.critical(log_message)



from .simple_module import SimpleModule
from .message import Message
from .compound_module import CompoundModule
from .channel import Channel

__all__ = ["SimpleModule", "Message", "CompoundModule", "Channel"]


"""This module implements the connector to the SimPy simulation engine."""
from omnetpypy.backends.connector import Connector
import simpy


class SimPyConnector(Connector):

    def __init__(self, simulation, metrics=None, output_dir=None, repetition=0):
        super().__init__(simulation, metrics, output_dir, repetition)
        self.env = simpy.Environment()
        self.entities = {}
        self._wrappers = {}

    def start_simulation(self, until=None):
        # we load a process that calls initialize for each entity
        self.env.process(initialize_entity(self.simulation.network))
        if until:
            self.env.run(until=until)
        else:
            self.env.run()

    def add_entity(self, entity):
        # entity is a simple module
        # we need to implement the module as a process
        # and add it to the environment
        super().add_entity(entity)

        self.entities[entity.name] = entity

        if entity.is_listening:
            self._wrappers[entity.identifier] = ModuleProcessWrapper(entity, self.env)

    def get_time(self):
        return self.env.now

    def schedule_port_input(self, port, message):
        # put the message in the store of the module wrapper
        self._wrappers[port.parent.identifier].put(message, port.name)

    def schedule_self_message(self, message, entity, at=None, delay=None):
        if not entity.is_listening:
            raise Exception(f"Entity {entity} is not listening. Cannot schedule a self message.")
        # create a process using the schedule_message function
        process = self.env.process(schedule_message(message, self._wrappers[entity.identifier], at, delay))
        # put the message and the process in the scheduled messages list
        self._wrappers[entity.identifier].scheduled_messages.append((message, process))

    def is_scheduled(self, message, entity):
        # check if the message is in the scheduled messages list
        for m, p in self._wrappers[entity.identifier].scheduled_messages:
            if m == message:
                return True
        return False

    def cancel_scheduled(self, message, entity):
        # cancel the scheduled message
        for m, p in self._wrappers[entity.identifier].scheduled_messages:
            if m == message:
                p.interrupt()
                self._wrappers[entity.identifier].scheduled_messages.remove((m, p))


class ModuleProcessWrapper:

    def __init__(self, module, env):
        self.module = module
        # define a store for the messages
        self.store = simpy.Store(env)
        # define a process for the module
        self.process = env.process(self.run())

        self.env = env

        self.scheduled_messages = []

    def run(self):
        while True:
            try:
                message, port_name = yield self.store.get()
                self.module.handle_message(message, port_name)
            except simpy.Interrupt:
                break

    def put(self, message, port_name):
        self.store.put((message, port_name))


def schedule_message(message, wrapper, at=None, delay=None):
    # wait for a given time (either at or delay) and then put a message in the store
    try:
        if at:
            yield wrapper.env.timeout(at - wrapper.env.now)
        elif delay:
            yield wrapper.env.timeout(delay)
        wrapper.put(message=message, port_name=None)
        # remove the message from the scheduled messages list
        for m, p in wrapper.scheduled_messages:
            if m == message:
                wrapper.scheduled_messages.remove((m, p))
    except simpy.Interrupt:
        pass


def initialize_entity(entity):
    for step in range(0, 6):
        yield entity.sim_context.connector.env.timeout(0)
        initialize_entity_with_step(entity, step)


def initialize_entity_with_step(entity, step):
    entity.initialize(step)
    if hasattr(entity, "sub_entities"):
        for sub_entity in entity.sub_entities.values():
            initialize_entity_with_step(sub_entity, step)



"""
This module implements the Message class, representing a message in the simulation.
"""

__all__ = ['Message']


class Message:
    r"""
    This class represents the messages exchanged between modules in a simulation.
    """

    def __init__(self, fields, **meta):
        self.fields = fields
        self.meta = meta
        if meta is None:
            self.meta = {}

    def __str__(self):
        return f"Message(fields={self.fields}, meta={self.meta})"

    def __copy__(self):
        new_meta = self.meta.copy()
        return Message(fields=self.fields[:], **new_meta)



"""
This module implements the Port class, which has the same semantic as in omnet++, except that
every port is an object and can does not have to be input or output only.
"""


class Port:

    def __init__(self, name, parent):
        self.name = name
        self.connected_port = None
        self.forwarded_input_port = None
        self.forwarded_output_port = None
        self.parent = parent
        self.subscribed_ports = []
        self.is_subscribed = False

    def connect(self, port):
        if self.is_subscribed:
            raise ValueError("Cannot connect a port that is already subscribed to another port")
        if port.connected_port is not None:
            raise ValueError("Cannot connect a port that is already connected")
        if self.forwarded_output_port:
            raise ValueError("Cannot connect a port that is forwarding output")

        self.connected_port = port
        port.connected_port = self

    def subscribe_to(self, port):
        if self.connected_port is not None:
            raise ValueError("Cannot subscribe a port that is already connected")
        if port.forwarded_output_port:
            raise ValueError("Cannot subscribe to a port that is forwarding output")
        if port.connected_port is not None:
            raise ValueError("Cannot subscribe to a port that is already connected")

        port.subscribed_ports.append(self)
        self.is_subscribed = True

    def disconnect(self, port):
        self.connected_port = None
        port.connected_port = None

    def forward_input(self, port):
        self.forwarded_input_port = port

    def forward_output(self, port):
        self.forwarded_output_port = port

    def tx_output(self, message):

        if self.forwarded_output_port:
            self.forwarded_output_port.tx_output(message)
        elif self.connected_port:
            self.connected_port.tx_input(message)
        elif len(self.subscribed_ports) > 0:
            for port in self.subscribed_ports:
                message_cpy = message.__copy__()
                port.tx_input(message_cpy)
        else:
            # do nothing, the mesaage is lost
            pass

    def tx_input(self, message):

        if self.forwarded_input_port:
            self.forwarded_input_port.tx_input(message)
        elif self.parent.is_listening:
            self.parent.sim_context.connector.schedule_port_input(self, message)
        else:
            # do nothing, the mesaage is lost
            pass



from omnetpypy.front_end import SimpleModule, Message, CompoundModule, Channel
from omnetpypy.simulation import Experiment

__all__ = ["SimpleModule", "Message", "CompoundModule", "Channel", "Experiment"]



from omnetpypy.front_end.port import Port


class SimulatedEntity:
    """
    A simulated entity is any active entity belonging to the simulation, such as modules and channels.
    Every entity may have a set of ports, and must keep a reference to the simulation context.
    """

    def __init__(self, name, identifier, port_names):
        self.name = name
        self.identifier = identifier
        self.sim_context = None
        self.parent = None

        self.is_listening = False
        # if the entity is listening, the connector will call the handle_message method when a message is received

        self.ports = {port_name: Port(port_name, parent=self) for port_name in port_names}

    def set_sim_context(self, sim_context):
        self.sim_context = sim_context

    def handle_message(self, message, port_name):
        """
        Process a message received as input to a port.

        Parameters
        ----------
        message : Message
            The message to be processed.
        port_name : Port
            The port from which the message was received.
        """
        pass

    def schedule_message(self, message, at=None, delay=None):
        # this sends a self message to the module at the specified time
        self.sim_context.connector.schedule_self_message(entity=self, message=message, at=at, delay=delay)

    def is_scheduled(self, message):
        return self.sim_context.connector.is_scheduled(message, self)

    def send(self, message, port_name):
        self.ports[port_name].tx_output(message)

    def cancel_scheduled(self, message):
        self.sim_context.connector.cancel_scheduled(message, self)

    def initialize(self, step=0):
        pass


from setuptools import setup, find_packages

setup(
    name='omnetpypy',
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        # Add your dependencies here
    ],
    entry_points={},
    package_data={
        'omnetpypy': ['data/seeds.csv'],
    },
    include_package_data=True,
    url='https://github.com/LBacciottini/OmnetPyPy',
    author='Leonardo Bacciottini',
    author_email='baccio.leonardo@gmail.com',
    description='An omnet++-like simulation framework fully implemented on Python, '
                'with the possibility to support different simulation backends such as SimPy.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='MIT',
)



"""
This class implements the SimpleModule class, with the same semantic as in omnet++
"""
from abc import abstractmethod

from omnetpypy.front_end.sim_entity import SimulatedEntity


class SimpleModule(SimulatedEntity):

    def __init__(self, name, identifier, port_names):
        super().__init__(name, identifier, port_names)
        self.is_listening = True

    @abstractmethod
    def handle_message(self, message, port_name):
        pass

    def connect(self, local_port, remote_entity, remote_port, channel=None):
        """
        Connect a port of this module to a port of another remote entity.

        Parameters
        ----------
        local_port : str
            The name of the local port to be connected.
        remote_entity : SimulatedEntity
            The remote entity to which the port will be connected.
        remote_port : str
            The name of the remote port to be connected.
        channel : Channel or None, optional
            The channel through which the connection will be made, if any.
            If set, the local port is connected to the "A" port of the channel,
            and the "B" port of the channel is connected to the remote port.
        """

        if channel is not None:
            self.ports[local_port].connect(channel.ports["A"])
            channel.ports["B"].connect(remote_entity.ports[remote_port])
        else:
            self.ports[local_port].connect(remote_entity.ports[remote_port])

    def emit_metric(self, name, value):
        """
        Record a metric sample in the simulation context.

        Parameters
        ----------
        name : str
            The name of the metric
        value
            The value of the metric sample
        """
        self.sim_context.connector.record_metric(name, value)



"""
This module implements the CompoundModule class, with the same semantic as in omnet++
"""
from omnetpypy.front_end.sim_entity import SimulatedEntity


class CompoundModule(SimulatedEntity):

    def __init__(self, name, identifier, port_names, **kwargs):
        super().__init__(name, identifier, port_names)

        # additional attributes based on kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.sub_entities = {}

    def add_sub_entity(self, module):
        if module.name in self.sub_entities:
            raise Exception("A submodule with the same name already exists")
        self.sub_entities[module.name] = module
        module.parent = self

    def get_sub_entity(self, name):
        # SHOULD NEVER BE USED IN A SIMULATION
        return self.sub_entities[name]

    def forward_input(self, port_container, port_submodule):
        self.ports[port_container].forward_input(port_submodule)

    def forward_output(self, port_submodule, port_container):
        self.ports[port_submodule].forward_output(port_container)

    def connect(self, local_port, remote_entity, remote_port, channel=None):
        """
        Connect a port of this module to a port of another remote entity.

        Parameters
        ----------
        local_port : str
            The name of the local port to be connected.
        remote_entity : SimulatedEntity
            The remote entity to which the port will be connected.
        remote_port : str
            The name of the remote port to be connected.
        channel : Channel or None, optional
            The channel through which the connection will be made, if any.
            If set, the local port is connected to the "A" port of the channel,
            and the "B" port of the channel is connected to the remote port.
        """

        if channel is not None:
            self.ports[local_port].connect(channel.ports["A"])
            channel.ports["B"].connect(remote_entity.ports[remote_port])
        else:
            self.ports[local_port].connect(remote_entity.ports[remote_port])

    def schedule_message(self, message, at=None, delay=None):
        raise Exception("CompoundModule does not support scheduling messages")

    def handle_message(self, message, port_name):
        raise Exception("CompoundModule does not support handling messages")

    def send(self, message, port_name):
        raise Exception("CompoundModule does not support sending messages")

    def cancel_scheduled(self, message):
        raise Exception("CompoundModule does not support cancelling messages")

    def is_scheduled(self, message):
        raise Exception("CompoundModule does not support checking if a message is scheduled")


"""
This module implements the Channel class, which implements advanced connectivity between modules in the simulation.
"""
from omnetpypy.front_end import Message
from omnetpypy.front_end.sim_entity import SimulatedEntity


class Channel(SimulatedEntity):

    def __init__(self, name, identifier=None, delay=None, loss_prob=None):
        super().__init__(name, identifier, port_names=["A", "B"])
        self.is_listening = True
        self.delay = delay
        self.loss_prob = loss_prob

    def handle_message(self, message, port_name):
        """
        Process a message received from a port. First, it processes the message, then it applies
        optional delay, and finally it sends the message out of the other port.

        Parameters
        ----------
        message : Message
            The message to be handled.
        port_name : str
            The port from which the message was received.
        """

        if port_name is None:  # this is a self message, should be immediately sent to the other port
            port = message.meta["port_noneshouldusethiskey"]
            del message.meta["port_noneshouldusethiskey"]
            self.send(message, port)
            return

        post_processed_message = self.process_message(message, port_name)

        should_drop = self.apply_loss(message, port_name) or post_processed_message is None
        out_port = "B" if port_name == "A" else "A"
        if not should_drop and (self.delay is None or self.delay <= 0):
            self.send(post_processed_message, out_port)
        elif not should_drop:
            post_processed_message.meta["port_noneshouldusethiskey"] = out_port
            self.schedule_message(post_processed_message, delay=self.generate_delay(message, port_name))

    def process_message(self, message, port_name):
        """
        Process a message received from a port.

        Parameters
        ----------
        message : Message
            The message to be processed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        Message or None
            The message to be sent to the connected port, or None if the message should be dropped for any reason.
        """

        # default behavior
        return message

    def generate_delay(self, message, port_name):
        """
        Generate a delay for the message, based on an arbitrary delay distribution.

        Parameters
        ----------
        message : Message
            The message to be delayed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        float or None
            The delay to be applied to the message. If None or zero, the message will be sent immediately.
        """

        # default behavior
        return self.delay

    def apply_loss(self, message, port_name):
        """
        Apply loss to the message, based on an arbitrary loss probability distribution.

        Parameters
        ----------
        message : Message
            The message to be delayed.
        port_name : str
            The port from which the message was received.

        Returns
        -------
        bool
            True if the message should be lost, False otherwise.
        """

        # default behavior
        if self.loss_prob is not None:

            # we use the default rng in the sim_context
            return self.sim_context.rng.random() < self.loss_prob

        return False

